package at.oefai.aaa.agent.jam.types;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;

/**
 * Represents a plan's variable bindings.
 * @author Marc Huber
 * @author Jaeho Lee
 * @author Stefan Rank
 */
public class Binding implements Serializable {
    // true if the binding is done with at least one new WM entry:
    private boolean newWMBinding = false;
    private SymbolMap symbolMap = null; // the mother map (where the variables are from)
    private Map<Variable, BindingValue> bvalues = new HashMap<Variable, BindingValue>();

    /** Constructor based on an externally-supplied symbol table. */
    public Binding(final SymbolMap pSymbolMap) {
        this.symbolMap = pSymbolMap;
    }

    /** Copy constructor (moderately deep copy: no duplication of external binds). */
    public Binding(final Binding b) {
        if (b != null) {
            this.symbolMap = b.symbolMap;
            this.newWMBinding = b.newWMBinding;
            // the following is not threadsafe (a 'deep' copy) throws
            // ConcurrentModificationException
            for (Map.Entry<Variable, BindingValue> e : b.bvalues.entrySet()) {
                this.bvalues.put(e.getKey(), new BindingValue(e.getValue()));
            }
        } else {
            this.symbolMap = null;
            this.newWMBinding = false;
        }
    }

    // Member functions

    /** 'Assignment' operator (shallow copy). */
    public void copy(final Binding b) {
        this.newWMBinding = b.isNewWMBinding();
        this.symbolMap = b.symbolMap;
        this.bvalues = b.bvalues;
    }

    /** Restore variable in the expression to an undefined state.
     * do nothing if the expression isnt a variable */
    public void unbindVariable(final Expression expression) {
        if (expression instanceof Variable) {
            this.bvalues.remove(expression);
        }
    }

    /** Restore variables in the expression list to an undefined state. */
    public void unbindVariables(final ExpList expressions) {
        for (Expression e : expressions) {
            this.unbindVariable(e);
        }
    }

    /** Tie this binding with an external variable binding. */
    public void linkVariables(final Variable var, final Variable extVariable, final Binding extBinding) {
        BindingValue bval = this.bvalues.get(var);
        if (bval == null) {
            bval = new BindingValue(extVariable, extBinding);
            this.bvalues.put(var, bval);
        } else {
            bval.setExternal(extVariable, extBinding);
        }
    }

    /** Set the internal value of the variable. */
    public void setValue(final Expression var, final Value val) {
        if (var instanceof Variable) {
            BindingValue bval = this.bvalues.get(var);
            if (bval == null) {
                this.bvalues.put((Variable) var, new BindingValue(val));
            } else if (bval.getExternalVariable() == null) {
                bval.setValue(val);
            } else {
                bval.getExternalBinding().setValue(bval.getExternalVariable(), val);
            }
        } else {
            assert false : "Binding value (" + val + ") to a non-variable (" + var + ") wont work";
        }
    }

    /** Get the variable's value by looking up the internal ID. */
    public Value getValue(final Variable variable) {
        BindingValue bval = this.bvalues.get(variable);
        if (bval == null) { // not bound
            return Value.UNDEFINED;
        } else if (bval.getExternalVariable() == null) {
            return bval.getValue();
        } else {
            return bval.getExternalBinding().getValue(bval.getExternalVariable());
        }
    }

    /** See if there are any references based upon a newly changed world model entry. */
    public boolean isNewWMBinding() {
        return this.newWMBinding;
    }

    /** Check, and possibly alter the flag indicating a reference to a newly changed world model entry. */
    public boolean checkNewWMBinding(final boolean newWM) {
        return this.newWMBinding = this.newWMBinding || newWM;
    }

    /** Indicate that there are no references to newly changed world model entries. */
    public void clearNewWMBinding() {
        this.newWMBinding = false;
    }

    /** Print out without worrying about being in-line with other output. */
    public String verboseString() {
        StringBuffer sb = new StringBuffer("  Binding:\n");
        sb.append("    Size:        \t" + this.bvalues.size() + "\n");
        sb.append("    NewWMBinding:\t" + this.newWMBinding + "\n");
        // iterate over all variables in the map
        //try { // why was this here ???
        for (Map.Entry<Variable, BindingValue> e : this.bvalues.entrySet()) {
            Variable var = e.getKey();
            BindingValue bval = e.getValue();
            sb.append("      [" + var.getName() + " = ");
            if (bval.getExternalVariable() == null) {
                sb.append("ExternalVar: " + bval.getExternalVariable().getName());
                sb.append("]\n");
            } else {
                sb.append(bval.getValue().verboseString(this));
                sb.append("]\n");
            }
        }
        //} catch(Exception ex) {
        //    sb.append("Interrupted by " + ex.toString());
        //}
        sb.append("\n");
        return sb.toString();
    }

    /** Print out so that it can be in-line with other output. */
    public String formattedString() {
        StringBuffer sb = new StringBuffer();
        sb.append("  Binding: size=" + this.bvalues.size() + ", new=" + this.newWMBinding);
        //try { // why was this here ???
        for (Map.Entry<Variable, BindingValue> e : this.bvalues.entrySet()) {
            sb.append(", ");
            Variable var = e.getKey();
            BindingValue bval = e.getValue();
            sb.append("[" + var.getName() + " = ");
            if (bval.getExternalVariable() != null) {
                sb.append("ExternalVar: " + bval.getExternalVariable().getName());
                sb.append("]");
            } else {
                sb.append(bval.getValue().verboseString(this));
                sb.append("]");
            }
        }
        //} catch(Exception ex) {
        //    sb.append("Interrupted by " + ex.toString());
        //}
        return sb.toString();
    }

}
