package at.oefai.aaa.agent.jam.types;

import java.io.Serializable;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.NoSuchElementException;

/**
 * DList a wrapper for the standard LinkedList.
 * @author Stefan Rank
 */
public class DList<E> implements Serializable, Iterable<E> {
    private static final DList EMPTY_DLIST = new EmptyDList();
    private final LinkedList<E> list;

    /** Explicit default constructor. */
    public DList() {
        super();
        this.list = new LinkedList<E>();
    }

    /** ShallowCoy constructor. */
    public DList(final DList<E> pList) {
        this();
        this.list.addAll(pList.list);
    }


    /** New one element list. */
    public DList(final E ent) {
        this();
        this.list.addFirst(ent);
    }

    /** Clear the whole list. */
    public void clear() {
        this.list.clear();
    }

    /** Set the nth (1-based) element of the list. */
    public E setNth(final int n, final E ent) {
        return this.list.set(n - 1, ent);
    }

    /** Remove the nth (1-based) element of the list. */
    public E removeNth(final int n) {
        return this.list.remove(n - 1);
    }

    /** Return the nth (1-based) element of the list. */
    public E getNth(final int n) {
        return this.list.get(n - 1);
    }

    /** Needed to use DList in a for-each loop.
     * @see java.lang.Iterable#iterator()
     */
    public Iterator<E> iterator() {
        return this.list.iterator();
    }

    public ListIterator<E> listIterator() {
        return this.list.listIterator();
    }

    public int size() {
        return this.list.size();
    }

    public void addLast(final E o) {
        this.list.addLast(o);
    }

    public boolean add(final E o) {
        this.list.addLast(o);
        return true;
    }

    public void addFirst(final E o) {
        this.list.addFirst(o);
    }

    public E getFirst() {
        return this.list.getFirst();
    }

    public E getLast() {
        return this.list.getLast();
    }

    public E removeFirst() {
        return this.list.removeFirst();
    }

    public E removeLast() {
        return this.list.removeLast();
    }

    public void shuffle() {
        Collections.shuffle(this.list);
    }

    public void sort(final Comparator<E> c) {
        Collections.sort(this.list, c);
    }

    public int indexOf(final E o) {
        return this.list.indexOf(o);
    }

    public List<E> shallowListClone() {
        return (List<E>) this.list.clone();
    }

    public static final <E> DList<E> emptyDList() {
        return (DList<E>) EMPTY_DLIST;
    }

    /** An unmodifiable empty object list. */
    private static class EmptyDList extends DList<Object> {
        public void clear() { }

        public Object setNth(final int n, final Object ent) {
            throw new IndexOutOfBoundsException();
        }

        public Object removeNth(final int n) {
            throw new IndexOutOfBoundsException();
        }

        public Object getNth(final int n) {
            throw new IndexOutOfBoundsException();
        }

        public ListIterator<Object> listIterator() {
            return new ListIterator<Object>() {
                public boolean hasNext() { return false; }
                public boolean hasPrevious() { return false; }
                public Object next() { throw new NoSuchElementException(); }
                public Object previous() { throw new NoSuchElementException(); }
                public int nextIndex() { return 0; }
                public int previousIndex() { return -1; }
                public void remove() { throw new UnsupportedOperationException(); }
                public void set(final Object o) { throw new UnsupportedOperationException(); }
                public void add(final Object o) { throw new UnsupportedOperationException(); }
            };
        }

        public int size() {
            return 0;
        }

        public void addLast(final Object o) {
            throw new UnsupportedOperationException();
        }

        public boolean add(final Object o) {
            throw new UnsupportedOperationException();
        }

        public void addFirst(final Object o) {
            throw new UnsupportedOperationException();
        }

        public Object getFirst() {
            throw new NoSuchElementException();
        }

        public Object getLast() {
            throw new NoSuchElementException();
        }

        public Object removeFirst() {
            throw new NoSuchElementException();
        }

        public Object removeLast() {
            throw new NoSuchElementException();
        }

        public void shuffle() {
        }

        public void sort(final Comparator<Object> c) {
        }

        public int indexOf(final Object o) {
            return -1;
        }

        public List<Object> shallowListClone() {
            return Collections.emptyList();
        }
    }
}
