package at.oefai.aaa.agent.jam.types;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;

/**
 * A list with fixed size, if an excess element gets added, the oldest one is discarded.
 * @author Stefan Rank
 */
public class FixedSizeList<E> implements Iterable<E>, Serializable {
    private final E[] elements;
    private final int capacity;
    protected int nextpos = 0; // protected for inner class access
    private int count = 0;

    public FixedSizeList(final int pCapacity) {
        if (pCapacity < 1) {
            throw new IllegalArgumentException();
        }
        this.capacity = pCapacity;
        this.elements = (E[]) new Object[pCapacity];
    }

    public int size() {
        return this.count;
    }

    /** Protected so the inner class has access. May return null.
     * @throws IndexOutOfBoundsException if index is out of range (index<0 || index>=capacity)
     * @throws NoSuchElementException if index points to null entry
     */
    protected E get(final int index) {
        if (index >= this.capacity || index < 0) {
            throw new IndexOutOfBoundsException("Index: " + index + ", Capacity: " + this.capacity);
        }
        E result = this.elements[index];
        if (result == null) {
            throw new NoSuchElementException();
        }
        return result;
    }

    /** Used in inner class Iterator. Searches from start (exclusive) for non-null elements.
     * @throws NoSuchElementException if list is empty (or maybe concurrentmodification)
     * @return an index
     */
    protected int nextNonNullIndex(final int start) {
        int pos = start;
        // use bounded for (instead of while) to avoid endless loops in case of concurrent delete of last element
        for (int i = 0; i < this.capacity; ++i) {
            pos = (pos + 1) % this.capacity;
            if (this.elements[pos] != null) {
                return pos;
            }
        }
        throw new NoSuchElementException(); // could also be a ConcurrentModification
    }

    public boolean add(final E element) {
        // allow adding null elements
        if (element != null) {
            if (this.elements[this.nextpos] == null) { // a new one is added
                ++this.count;
            }
        } else {
            if (this.elements[this.nextpos] != null) { // the oldest one is erased
                --this.count;
            }
        }
        this.elements[this.nextpos] = element;
        this.nextpos = (this.nextpos + 1) % this.capacity;
        return true;
    }

    public void remove(final List<Integer> indices) {
        assert (indices != null);
        if (indices.size() == 0) { return; }
        for (int i : indices) {
            if (this.elements[i] != null) {
                this.elements[i] = null;
                --this.count;
            }
        }
    }

    /** Return index of equal object or -1 if not found. */
    public int indexOf(final E element) {
        assert (element != null) : "non-null argument required";
        for (int i = 0; i < this.capacity; ++i) {
            if (element.equals(this.elements[i])) {
                return i;
            }
        }
        return -1;
    }

    public List<Integer> indicesOf(final E element) {
        assert (element != null) : "non-null argument required";
        List<Integer> indices = new ArrayList<Integer>();
        for (int i = 0; i < this.capacity; ++i) {
            if (element.equals(this.elements[i])) {
                indices.add(i);
            }
        }
        return indices;
    }

    /** Simple Iterator that does not allow modification.
     *  guarantees ordering of the returned elements as oldest first.
     */
    public Iterator<E> iterator() {
        return new Itr();
    }

    /** Inner Iterator implementation, does not allow modification. */
    private class Itr implements Iterator<E> {
        /** Index of element-search-start for subsequent call to next. */
        private int nextIndex = nextpos - 1;
        /** Count of already returned elements. */
        private int returnCount = 0;

        public boolean hasNext() {
            return this.returnCount != size();
        }

        /** Return next element, the array access automatically throws an exception. */
        public E next() {
            try {
                nextIndex = nextNonNullIndex(nextIndex);
                E next = get(nextIndex);
                ++this.returnCount;
                return next;
            } catch (IndexOutOfBoundsException e) {
                throw new NoSuchElementException();
            }
        }

        public void remove() {
            throw new UnsupportedOperationException();
        }
    }

}
