package at.oefai.aaa.agent.jam.types;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * IndexedMap a simple combination of a mapping and an iterate-able listing of the mapped elements
 * String keys are mapped to possibly multiple elements, all entries are accessible by index.
 * all entries for a specific key can be returned as DList
 * @author Stefan Rank
 */
public class IndexedMap<K,V> implements Serializable {
    private final Map<K,DList<V>> map = new HashMap<K,DList<V>>(32);
    private final ArrayList<V> indexedList = new ArrayList<V>(); // contains all elements

    public int size() {
        return this.indexedList.size();
    }

    /** Retrieve the object referenced by the specified index. */
    public V get(final int index) {
        return this.indexedList.get(index);
    }

    public int indexOf(final V obj) {
        return this.indexedList.indexOf(obj);
    }

    /**
     * Add an entry into the list for the specified key.
     * @param key the string key this object belongs to
     * @param obj the object to add
     * @return true if this map changed as a result of the call
     */
    public boolean put(final K key, final V obj) {
        DList<V> maplist = this.map.get(key);
        if (maplist != null) { // already a list for this key
            int index = maplist.indexOf(obj);
            if (index > -1) { // already in hashmap list
                return false; // didnt change
            }
            // new one
            maplist.add(obj);
        } else { // need to create a new DList
            maplist = new DList<V>(obj);
            this.map.put(key, maplist);
        }
        // now into the indexedlist:
        this.indexedList.add(obj);
        return true;
    }

    /**
     * Removes the object from the list for the key if present.
     * @param key the key to the lists
     * @param obj the object to remove
     * @return true if the list changed
     */
    public boolean remove(final K key, final V obj) {
        DList<V> maplist = this.map.get(key);
        if (maplist != null) { // already a list for this key
            int index = maplist.indexOf(obj);
            if (index > -1) { // it is in hashmap list, so remove
                maplist.removeNth(index + 1);
                // and remove from indexed list
                this.indexedList.remove(obj);
                return true;
            }
            // not there
            return false;
        }
        // no list so there can be no object
        return false;
    }


    /** Return the entire list of objects in the table that match the specified label.
     *  or an unmodifiable empty list if none matches. */
    public DList<V> getBucket(final K key) {
        DList<V> result = this.map.get(key);
        if (result != null) {
            return result;
        } else {
            return DList.emptyDList();
        }
    }

    public List<V> asList() {
        return this.indexedList;
    }
}
