package at.oefai.aaa.stage;

import java.io.Serializable;
import java.util.Map;
import java.util.concurrent.CountDownLatch;

import at.oefai.aaa.agent.Environment;
import at.oefai.aaa.agent.SchundActor;
import at.oefai.aaa.thread.IPausable;


/**
 * Contains the graphics (i.e. the environment presentation) and the displayed agents.
 * The basic setup for this object is described by the .aaastage files and there should
 * be a way to store snapshots of this object when the stage-action is paused. (TODO)
 * @author Stefan Rank
 */
public class ActingStage implements IPausable, Serializable {
    private Environment environment = null;
    private IPausable animEngine = null;
    private Map<String,SchundActor> agents = null;

    public ActingStage(final Environment pEnv, final Map<String,SchundActor> pAgents,
                       final IPausable pEngine) throws StageException {
        // check that arguments are non-null and the set contains SchundAgents
        if (pEnv == null || pAgents == null || pEngine == null) {
            throw new StageException("null arguments in ActingStage-constructor");
        }
        this.environment = pEnv;
        this.agents = pAgents;
        this.animEngine = pEngine;
        // check if agents are IPausable - not necessary any more, thanks to generics support
    }

    public final void start(final CountDownLatch startSignal) {
        // a startSignal can simply be passed down
        this.animEngine.start(null); // I think this should never be delayed
        // start the environment then the agents
        this.environment.start(startSignal);
        for (SchundActor sa : this.agents.values()) {
            sa.start(startSignal);
        }
    }

    public final void pause(final CountDownLatch returnSignal) {
        final CountDownLatch localSignal; // final to allow use in a Thread below
        if (returnSignal != null) {
            // since a Latch is initialized with a count we have to use a new intermediary one
            localSignal = new CountDownLatch(this.agents.size() + 1 + 1);
            (new Thread() { // asynchronously wait for the returnSignal of our children
                public void run() {
                    try {
                        localSignal.await();
                    } catch (InterruptedException e) {
                        // ignore
                    }
                    returnSignal.countDown(); // now propagate the returnSignal upwards
                }
            }).start();
        } else {
            localSignal = null;
        }
        // pause the agents then the environment
        for (SchundActor sa : this.agents.values()) {
            sa.pause(localSignal);
        }
        this.environment.pause(localSignal);
        this.animEngine.pause(localSignal);
    }

    /**
     * if any composition object is still running return false.
     * @return state of the whole true means all paused
     */
    public final boolean isPaused() {
        if (! this.animEngine.isPaused()) { return false; }
        if (! this.environment.isPaused()) { return false; }
        for (SchundActor sa : this.agents.values()) {
            if (! sa.isPaused()) { return false; }
        }
        return true;
    }

    public final void synchronizeGraphics() {
        this.environment.execObserver();
    }

}
