package at.oefai.aaa.stage;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import at.oefai.aaa.IDetailSink;
import at.oefai.aaa.agent.ActorInterpreter;
import at.oefai.aaa.agent.Environment;
import at.oefai.aaa.agent.EnvironmentInterpreter;
import at.oefai.aaa.agent.SchundActor;
import at.oefai.aaa.agent.jam.Animator;
import at.oefai.aaa.agent.jam.IEnvironment;
import at.oefai.aaa.agent.jam.ParseException;
import at.oefai.aaa.agent.jam.TokenMgrError;
import at.oefai.aaa.animation.AnimationEngine;

/**
 * Factory for ActingStages, parses an xml definition file that points to other files.
 * @author Stefan Rank
 */
public class StageFactory {
    private ActingStage stage = null;

    public StageFactory(final File stageFile, final AnimationEngine animEng, final IDetailSink detailSink) throws StageException {
        assert (stageFile != null) : "non-null stageFile argument required";
        assert (animEng != null) : "non-null animEngine argument required";
        assert (detailSink != null) : "non-null detailSink argument required";
        Document stageDoc = getXMLDocument(stageFile);
        String stageFileDir = stageFile.getParent();
        // set the default loglevel if there is one
        setDefaultLogLevel(stageDoc);
        // first the graphics file
        File graphicsFile = getCheckedFile(stageFileDir, getVisualFilePath(stageDoc));

        try {
            animEng.setDocument(graphicsFile);
        } catch (IOException ex) {
            throw new StageException("parsing of " + graphicsFile + " failed\n", ex);
        }

        // before registering sounds, remove the old ones
        animEng.clearSounds();
        String soundDirPath = getSoundDirPath(stageDoc);
        if (soundDirPath != null) {
            File sndDirectory = getCheckedDir(stageFileDir, soundDirPath);
            registerSounds(sndDirectory, animEng);
        }
        //clear the sink for the detail display
        detailSink.clearModels();
        // then the environment
        Environment env = getEnvironment(stageDoc, stageFileDir, animEng, detailSink);
        // then the agents
        Map<String,SchundActor> agentMap = getActors(stageDoc, stageFileDir, env, detailSink, animEng);

        env.assertObjects(animEng.getObjectNames());

        env.assertAgents(agentMap.keySet());

        this.stage = new ActingStage(env, agentMap, animEng);
        // now start to show the document
        // it is now the responsibility of the animEngine
        animEng.showDocument();
    }

    public ActingStage getActingStage() {
        return this.stage;
    }


    private String getVisualFilePath(final Document stageDoc) throws StageException {
        // first the graphics file
        NodeList graphics = stageDoc.getElementsByTagName("visualFile");
        if (graphics.getLength() != 1) {
            throw new StageException("not exactly one visualFile element in stageFile");
        }
        String graphicsFilePath = ((Element) graphics.item(0)).getAttribute("location").trim();
        if (graphicsFilePath.equals("")) {
            throw new StageException("no visualFile location in stageFile");
        }
        return graphicsFilePath;
    }


    private String getSoundDirPath(final Document doc) throws StageException {
        NodeList snd = doc.getElementsByTagName("sounds");
        if (snd.getLength() < 1) {
            return null;
        }
        if (snd.getLength() > 1) {
            throw new StageException("more than one sound-dir element in stageFile");
        }
        String sndPath = ((Element) snd.item(0)).getAttribute("location").trim();
        if (sndPath.equals("")) {
            throw new StageException("empty sound directory pathstring found in stageFile");
        }
        return sndPath;
    }

    private void registerSounds(final File sndDir, final AnimationEngine eng) {
        File[] wavs = sndDir.listFiles(new FilenameFilter() {
            public boolean accept(final File f, final String name) {
                return name.endsWith(".wav");
            }
        });
        if (wavs != null) {
            for (int i = 0; i < wavs.length; ++i) {
                eng.registerSound(wavs[i]);
            }
        }
    }


    private File getCheckedFile(final String parent, final String filepath) throws StageException {
        File f = new File(filepath);
        if (! f.isAbsolute()) {
            f = new File(parent, f.getPath());
        }
        if (! (f.exists() && f.isFile())) {
            throw new StageException("location " + filepath + " isnt a file");
        }
        return f;
    }

    private File getCheckedDir(final String parent, final String dirpath) throws StageException {
        File f = new File(dirpath);
        if (! f.isAbsolute()) {
            f = new File(parent, f.getPath());
        }
        if (! (f.exists() && f.isDirectory())) {
            throw new StageException("location " + dirpath + " isnt a directory");
        }
        return f;
    }


    private void setDefaultLogLevel(final Document stageDoc) {
        String level = stageDoc.getDocumentElement().getAttribute("loglevel").trim().toUpperCase();
        if (! level.equals("")) {
            // try to set the loglevel of the root logger
            Logger log = Logger.getLogger("");
            try {
                Level newLevel = Level.parse(level);
                log.setLevel(newLevel);
                log.logp(Level.WARNING, null, null,
                         "StageFactory set root loglevel to " + level);
            } catch (IllegalArgumentException ex) {
                log.logp(Level.WARNING, null, null,
                         level + " (specified in stageFile) is not a valid loglevel,\n"
                         + "StageFactory could not set root logger's log level\n"
                         + "exception: " + ex.toString());
            } catch (SecurityException ex) {
                log.logp(Level.WARNING, null, null,
                         "StageFactory could not set root logger's log level\n"
                         + "wanted level (specified in stageFile): " + level + "\n"
                         + "SecurityException: " + ex.toString());
            }
        }
    }

    private Environment getEnvironment(final Document stageDoc, final String parentDir,
                                       final AnimationEngine ae, final IDetailSink sink) throws StageException {
        NodeList envs = stageDoc.getElementsByTagName("environment");
        if (envs.getLength() != 1) {
            throw new StageException("not exactly one environment element in stageFile");
        }
        String jamFile = ((Element) envs.item(0)).getAttribute("jamFile").trim();
        if (jamFile.equals("")) {
            throw new StageException("no environment jamFile location in stageFile");
        }
        File f = getCheckedFile(parentDir, jamFile);

        EnvironmentInterpreter ri = createEnvironmentInterpreter(f.getAbsolutePath(), ae);
        sink.setEnvironmentModels(ri.getWorldModelListModel(), ri.getIntentionTreeModel(),
                                  ri.getIntentionTreeSelectionModel());
        return new Environment(ri);
    }

    private EnvironmentInterpreter createEnvironmentInterpreter(final String file, final Animator anim) throws StageException {
        try {
            return new EnvironmentInterpreter("Environment", new String[] {file}, anim);
        } catch (IOException ex) {
            throw new StageException("IOException for ReactiveInterpreter (" + file + ")", ex);
        } catch (ParseException ex) {
            throw new StageException("ParseException for ReactiveInterpreter (" + file + ")", ex);
        } catch (TokenMgrError error) {
            throw new StageException("TokenMgrError for ReactiveInterpreter (" + file + ")", error);
        }
    }


    private Map<String,SchundActor> getActors(final Document stageDoc, final String parentDir,
                                              final Environment env, final IDetailSink sink,
                                              final AnimationEngine animEng) throws StageException {
        Map<String,SchundActor> agentMap = new HashMap<String,SchundActor>(2 * 2 * 2);
        // get the basicPlans for all agents first:
        List<String> basicPlanFiles = new ArrayList<String>();
        NodeList basicnodes = stageDoc.getElementsByTagName("basicPlans");
        for (int i = 0; i < basicnodes.getLength(); ++i) {
            // get the jamfile:
            String jamFile = ((Element) basicnodes.item(i)).getAttribute("jamFile").trim();
            if (jamFile.equals("")) {
                throw new StageException("empty jamFile location for basicPlans");
            }
            basicPlanFiles.add(getCheckedFile(parentDir, jamFile).getAbsolutePath());
        }
        NodeList nodes = stageDoc.getElementsByTagName("agent");
        for (int i = 0; i < nodes.getLength(); i++) {
            String agentName = ((Element) nodes.item(i)).getAttribute(
                    "name").trim();
            if (agentName.equals("")) {
                throw new StageException("empty name for an agent");
            }
            // get the jamfile:
            String jamFile = ((Element) nodes.item(i)).getAttribute(
                           "jamFile").trim();
            if (jamFile.equals("")) {
                throw new StageException("empty jamFile location for agent (name: " + agentName + ")");
            }
            File jamf = getCheckedFile(parentDir, jamFile);
            // get the graphicsfile:
            String graphicsFile = ((Element) nodes.item(i)).getAttribute("svgFile").trim();
            if (graphicsFile.equals("")) {
                throw new StageException("empty svgFile location for agent (name: " + agentName + ")");
            }
            File svgf = getCheckedFile(parentDir, graphicsFile);
            // get the pngfile:
            String pngFile = ((Element) nodes.item(i)).getAttribute("pngFile").trim();
            if (pngFile.equals("")) {
                throw new StageException("empty pngFile location for agent (name: " + agentName + ")");
            }
            File pngf = getCheckedFile(parentDir, pngFile);
            // now append the agents svg doc
            try {
                animEng.appendAgentToDocument(svgf, agentName, i);
            } catch (IOException ex) {
                throw new StageException("doc-parsing of " + svgf + " for agent " + agentName + " failed\n", ex);
            }
            // create the new agent
            ActorInterpreter ai = createActorInterpreter(agentName, jamf.getAbsolutePath(), basicPlanFiles, env);
            SchundActor agent =  new SchundActor(agentName, ai);
            // also pass the svg file uri to the gui
            sink.addAgentModels(agentName, pngf, ai.getWorldModelListModel(),
                                ai.getIntentionTreeModel(), ai.getIntentionTreeSelectionModel());
            agentMap.put(agentName, agent);
        }
        return agentMap;
    }

    private ActorInterpreter createActorInterpreter(final String name, final String file,
            final List<String> basicPlanFiles, final IEnvironment env) throws StageException {
        try {
            int size = basicPlanFiles.size() + 1;
            String[] sa = new String[size];
            sa[0] = file;
            for (int i = 1; i < size; ++i) {
                sa[i] = basicPlanFiles.get(i - 1);
            }
            return new ActorInterpreter(name, sa, env);
        } catch (IOException ex) {
            throw new StageException("IOException for StoppableInterpreter (" + name + ")", ex);
        } catch (ParseException ex) {
            throw new StageException("ParseException for StoppableInterpreter (" + name + ")", ex);
        } catch (TokenMgrError error) {
            throw new StageException("TokenMgrError for ReactiveInterpreter (" + name + ")", error);
        }

    }


    /**
     * Helper method to encapsulate xml factory stuff.
     * @param f the xml file assumed non-null
     * @return the DOM Document
     * @throws StageLoadException with cause
     */
    private Document getXMLDocument(final File f) throws StageException {
        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
        dbf.setNamespaceAware(false);
        dbf.setValidating(false);
        DocumentBuilder db = null;
        Document doc = null;
        try {
            db = dbf.newDocumentBuilder();
            doc = db.parse(f); // parsing without validation as there is no schema/DTD (yet?)
        } catch (ParserConfigurationException ex) {
            throw new StageException("creating xml-parser for " + f + " failed", ex);
        } catch (IOException ex1) {
            throw new StageException("parsing of " + f + " failed", ex1);
        } catch (SAXException ex2) {
            throw new StageException("parsing of " + f + " failed", ex2);
        }
        return doc;
    }

}
